package au.com.fax2;

import java.io.*;
import java.util.*;
import java.text.*;
import java.nio.charset.Charset;

import org.apache.http.*;
import org.apache.http.auth.*;
import org.apache.http.client.*;
import org.apache.http.client.methods.*;
import org.apache.http.client.protocol.*;
import org.apache.http.entity.*;
import org.apache.http.impl.client.*;
import org.apache.http.message.*;

import com.google.gson.Gson;
import com.google.gson.JsonArray;
import com.google.gson.JsonElement;

public class Client {
	public static final String DEFAULT_BASE_URL = "https://api.fax2.com.au/v1";
	public static final String PROPERTIES_FILENAME = "fax2.properties";
	
	protected CloseableHttpClient httpClient;
	protected String accessToken;
	protected Gson gson;
	protected String default_username = null;
	protected String default_password = null;
	protected String base_url = DEFAULT_BASE_URL;
	
	protected long accessTokenExpiry;
	
	protected static SimpleDateFormat dateFormatter;
	
	static {
		dateFormatter = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ssX");
		dateFormatter.setTimeZone(TimeZone.getTimeZone("UTC"));
	}
	
	public static String dateAsISO8601(Date d) {
		return dateFormatter.format(d);
	}

	public class Exception extends java.lang.Exception {
		Exception(String message) {
			super(message);
		}
	}

	protected void init() {
		gson = new Gson();
		accessTokenExpiry = 0;
	}
	
	public Client(Properties properties) {
		init();
		default_username = properties.getProperty("fax2.username");
		default_password = properties.getProperty("fax2.password");
		base_url = properties.getProperty("fax2.base_url",DEFAULT_BASE_URL);
	}
	
	public Client() {
		init();
	}

	public class ApiResponse {
		String error, error_description;
	};

	class TokenResponse extends ApiResponse {
		String access_token, token_type;
		int expires_in;
	}
	
	class UploadResponse extends ApiResponse {
		String document_id;
		int pages;
	}
	
	public class SendFaxResponse extends ApiResponse {
		public String id;
		public int pages;
	}
	
	public class FaxStatus extends ApiResponse {
		public String status;
		public int pages;
		public Date sent_at;
		public String reason;
		public int send_attempts;
		public int pages_sent;
	}
	
	public class ReceivedFax {
		public String id;
		public String to;
		public Date received_at;
		public String service_id;
		public int pages;
	}
	
	public class ReceivedFaxesResponse extends ApiResponse {
		public JsonArray data;
		public String next_page_url;
	}
	
	public interface ReceivedFaxCollection extends Collection<ReceivedFax> {}
	
	public class ReceivedFaxList extends ArrayList<ReceivedFax> implements ReceivedFaxCollection {}
	
	protected void newAuthedClient() {
		Collection<Header> headers = new ArrayList<Header>();
		headers.add(new BasicHeader("Authorization","Bearer "+accessToken));
		httpClient = HttpClients.custom().setDefaultHeaders(headers).build();
	}
	
	protected void checkApiResponse(HttpResponse response) throws Exception {
		String content_type = response.getFirstHeader("Content-Type").getValue();
		String mime_type = content_type;
		int n = content_type.indexOf(';');
		if(n != -1) mime_type = content_type.substring(0,n);
		if(!mime_type.equals("application/json")) 
			throw new Exception("HTTP response does not have proper Content-Type of application/json");
	}
	
	protected void checkJsonResponse(ApiResponse apiResponse) {
		if(apiResponse == null) throw new Error("json didn't parse?");
		if(apiResponse.error != null)
			throw new Error(String.format("%s error: %s",apiResponse.error,apiResponse.error_description));
	}

	protected <T extends ApiResponse> T jsonRequest(HttpUriRequest req,Class<T> c)
		throws ClientProtocolException,IOException,Exception
	{
		if(accessTokenExpiry < new Date().getTime()) signIn();

		CloseableHttpResponse response = httpClient.execute(req);
		try {
			checkApiResponse(response);
			Reader json = new InputStreamReader(response.getEntity().getContent());
			T ret = gson.fromJson(json,c);
			checkJsonResponse(ret);
			return ret;
		} finally { response.close(); }
	}

	protected InputStream rawRequest(HttpUriRequest req)
		throws ClientProtocolException,IOException,Exception
	{
		if(accessTokenExpiry < new Date().getTime()) signIn();
		
		CloseableHttpResponse response = httpClient.execute(req);
		return response.getEntity().getContent();
	}

	protected void signIn(String username,String password)
		throws ClientProtocolException,IOException,Exception
	{
		CloseableHttpClient signinClient = HttpClients.createDefault();
		CredentialsProvider credsProvider = new BasicCredentialsProvider();
		credsProvider.setCredentials(
			new AuthScope(AuthScope.ANY),
			new UsernamePasswordCredentials(username,password)
		);
		HttpClientContext signinContext = HttpClientContext.create();
		signinContext.setCredentialsProvider(credsProvider);
		
		HttpUriRequest request = RequestBuilder.post(base_url+"/oauth2/token")
			.addParameter("grant_type","client_credentials")
			.build();

		CloseableHttpResponse response = signinClient.execute(request,signinContext);
		try {
			checkApiResponse(response);
			Reader json = new InputStreamReader(response.getEntity().getContent());
			TokenResponse tokenResponse = gson.fromJson(json,TokenResponse.class);
			checkJsonResponse(tokenResponse);
		
			accessToken = tokenResponse.access_token;
			newAuthedClient();
			accessTokenExpiry = new Date().getTime() +
				1000 * (tokenResponse.expires_in-60);
		} finally { response.close(); }
	}
	
	protected void signIn()
		throws ClientProtocolException,IOException,Exception
	{
		if(default_username == null)
			throw new Exception("Fax2 username has not been specified");
		if(default_password == null)
			throw new Exception("Fax2 password has not been specified");

		signIn(default_username,default_password);
	}
	
	public String addDocument(HttpEntity entity,String type)
		throws ClientProtocolException,IOException,Exception
	{
		HttpUriRequest request = RequestBuilder.post(base_url+"/upload_document")
			.setEntity(entity)
			.addHeader("Content-Type",type)
			.build();

		UploadResponse response = jsonRequest(request,UploadResponse.class);
		return response.document_id;
	}	
	
	public String addDocument(InputStream document,String type)
		throws ClientProtocolException,IOException,Exception
	{
		return addDocument(new InputStreamEntity(document),type);
	}
	
	public String addDocument(File file,String type)
		throws ClientProtocolException,IOException,Exception
	{
		return addDocument(new FileEntity(file),type);
	}
	
	protected static final Charset UTF8 = Charset.forName("UTF-8");

	public String addText(String text)
		throws ClientProtocolException,IOException,Exception
	{
		byte[] encoded = text.getBytes(UTF8);
		InputStream stream = new ByteArrayInputStream(encoded);
		return addDocument(stream,"text/plain; charset=utf-8");
	}
	
	public SendFaxResponse sendFax(String dest,Iterable<String> documents)
		throws ClientProtocolException,IOException,Exception
	{
		RequestBuilder builder = RequestBuilder.post(base_url+"/send_fax");
		builder.addParameter("dest_number",dest);
		for(String doc_id: documents) builder.addParameter("documents[]",doc_id);
		SendFaxResponse ret = jsonRequest(builder.build(),SendFaxResponse.class);
		return ret;
	}
	
	public FaxStatus getOutgoingFaxStatus(String fax_id)
		throws ClientProtocolException,IOException,Exception
	{
		HttpUriRequest request = RequestBuilder.get(base_url+"/sent_faxes/"+fax_id).build();

		return jsonRequest(request,FaxStatus.class);
	}
	
	public ReceivedFaxCollection getReceivedFaxes(Date from,Date before)
		throws ClientProtocolException,IOException,Exception
	{
		RequestBuilder builder;
		ReceivedFaxCollection ret = new ReceivedFaxList();
		
		builder = RequestBuilder.get(base_url + "/received_faxes");
		if(from != null)
			builder.addParameter("from_time",dateAsISO8601(from));
		if(before != null)
			builder.addParameter("before_time",dateAsISO8601(before));

		while(true) {
			ReceivedFaxesResponse resp = jsonRequest(builder.build(),ReceivedFaxesResponse.class);
			
			for(JsonElement el: resp.data) {
				ret.add(gson.fromJson(el,ReceivedFax.class));
			}
			if(resp.next_page_url == null) break;
			builder = RequestBuilder.get(resp.next_page_url);
		}
		return ret;
	}

	public ReceivedFaxCollection getReceivedFaxes(Date from)
		throws ClientProtocolException,IOException,Exception
	{
		return getReceivedFaxes(from,null);
	}

	public ReceivedFaxCollection getReceivedFaxes()
		throws ClientProtocolException,IOException,Exception
	{
		return getReceivedFaxes(null);
	}
	
	protected InputStream getReceivedFax(String id)
		throws ClientProtocolException,IOException,Exception
	{
		RequestBuilder builder = RequestBuilder.get(
			base_url + "/received_faxes/"+id+"/content.pdf"
		);

		return rawRequest(builder.build());
	}
	
	public void writeReceivedFax(String id,OutputStream out)
		throws ClientProtocolException,IOException,Exception
	{
		byte[] buf = new byte[16384];
		
		InputStream response = getReceivedFax(id);
		while(true) {
			int bytesread = response.read(buf);
			if(bytesread == -1) break;
			out.write(buf,0,bytesread);
		}
		response.close();
		return;
	}
	
	public void writeReceivedFax(String id,File outfile)
		throws ClientProtocolException,IOException,Exception
	{
		OutputStream out = new FileOutputStream(outfile);
		
		writeReceivedFax(id,out);
		out.flush();
		out.close();
	}
	
	public void writeReceivedFax(String id,String outfilename) 
		throws ClientProtocolException,IOException,Exception
	{
		writeReceivedFax(id,new File(outfilename));
	}
}
